#include <config.h>
program main_wloops
  use constants_module
  use comlib
  use lattice_class
  use error_class
  use timer_class
  use file_tools_class
  use field_gauge_class
  use config_profile_class
  use wilson_loop_class
  use hmc_stout_class
  use main_wloops_class
  implicit none
  type(config_profile_obj) :: config_prof
  type(vfield_gluon_wg), pointer :: u => NULL()
  type(lattice_world)  :: lattice
  type(WilsonLoopMeas) :: wloops
  type(stout_smearing_parameters) :: stout
  real(DP) :: stout_alpha
  integer  :: stout_nstep
  type(timer) :: etime,stout_etime
  real(DP) :: run_etime,smear_etime
  integer :: iflg,itraj,iout
  character(len=CHARLEN) :: fname, stmp
  character(len=CHARLEN) :: wl_param_fname,wl_out_path
  character(len=CHARLEN) :: config_path,config_name
  complex(DP) :: ue(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(DP) :: uo(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)

  call new(lattice)
  call new(etime)
  call tic(etime)

  iflg = 0
  if (nodeid == 0) then
    if ( 1 == COMMAND_ARGUMENT_COUNT() ) then
      call GET_COMMAND_ARGUMENT(1,fname)
    else
      iflg = 1
    endif
  endif
  if (NPU > 1) then
    call comlib_bcast(iflg,0)
    call comlib_bcast(fname,0)
  endif
  if ( 1 == iflg ) then
    call error_stop("Missing inputfile in command argument.")
  endif

  iout = search_free_file_unit()
  if ( 0 == nodeid ) then
    open(iout,file=TRIM(fname),status='old',form='formatted')
    read(iout,'(A)')stmp
    read(iout,*)itraj
    read(iout,'(A)')config_path
    read(iout,'(A)')config_name
    read(iout,'(A)')stmp
    read(iout,*)stout_alpha,stout_nstep
    read(iout,'(A)')wl_param_fname
    read(iout,'(A)')wl_out_path
    close(iout)
  endif

  if (NPU > 1) then
    call comlib_bcast(config_path,0)
    call comlib_bcast(config_name,0)
    call comlib_bcast(stout_alpha,0)
    call comlib_bcast(stout_nstep,0)
    call comlib_bcast(wl_param_fname,0)
    call comlib_bcast(wl_out_path,0)
  endif

  allocate(u)
  call new(u)
  call read_config(itraj,config_path,config_name,config_prof,u)

  if ( ABS(stout_alpha) > 0.0_DP .and. stout_nstep > 0 ) then
    call new(stout_etime)
    call tic(stout_etime)
    call new(stout,alpha=stout_alpha,nstout=stout_nstep,type=STOUT_TYPE_SPATIAL)
    call print(stout)
    call smear(stout,u)
    call toc(stout_etime)
    smear_etime = get_elapse(stout_etime)
    if ( 0 == nodeid ) then
      write(*,'(80("="))')
      write(*,'(" Smear ETime:",F14.7)')smear_etime
      write(*,'(80("="))')
    endif
    call convert(stout%usmr(stout_nstep),ue,uo)
    call delete(stout)
  else
    call convert(u,ue,uo)
  endif

  deallocate(u)
  u => NULL()

  call new(wloops,TRIM(wl_param_fname))

  if ( 0 == nodeid ) then
    write(*,'(80("="))')
  endif
  call write(wloops,6)
  if ( 0 == nodeid ) then
    write(*,'(80("="))')
  endif

  call run(wloops,TRIM(wl_out_path),itraj,ue,uo)

  call delete(wloops)

  call toc(etime)
  run_etime = get_elapse(etime)
  if ( 0 == nodeid ) then
    write(*,'(80("="))')
    write(*,'(" Total ETime:",F14.7)')run_etime
    write(*,'(80("="))')
  endif

  call delete(lattice)

  stop
end program
