static __device__ __inline__ void
cu_mult_clover_term(const cuclv_thread *fri,const int *ith, float *ytr, float *yti)
/*****************************************************
* Clover term mult (Chiral rep => Dirac rep.)
*  y := F y
*****************************************************/
{
  __align__(16) float pr[CSH],pi[CSH];
  __align__(16) float mr[CSH],mi[CSH];
//
//  p = y(up) + y(down)
//  m = y(up) - y(down)
//
#pragma unroll 6
  for (int ic=0; ic<CSH;ic++) {
    pr[ic] = ytr[ic] + ytr[ic+CSH];
    pi[ic] = yti[ic] + yti[ic+CSH];
    mr[ic] = ytr[ic] - ytr[ic+CSH];
    mi[ic] = yti[ic] - yti[ic+CSH];
  }
//
//  a = f2 * p
//  b = f1 * m
//
  clvvec1(fri,ith,mr,mi,ytr,yti);
  clvvec2(fri,ith,pr,pi,mr,mi);

//  y(1: 6) = a + b
//  y(7:12) = a - b
#pragma unroll 6
  for(int ic=0;ic<CSH;ic++){
    ytr[ic+CSH] = mr[ic] - ytr[ic];
    yti[ic+CSH] = mi[ic] - yti[ic];
    ytr[ic    ] = mr[ic] + ytr[ic];
    yti[ic    ] = mi[ic] + yti[ic];
  }
}
