#include <config.h>
module measure_class
  use lattice_class
  use error_class
  use file_tools_class
  use field_gauge_class
  use hmc_status_class
  use meson_class
  implicit none
  private
  public :: load_config
contains

subroutine load_config(path,head,status,u)
!
! Save configuration for measurment
!
! This is called in run mehtod and is private method.
!
  use file_tools_class
  use hmc_identity
  use print_status_class
  implicit none
  character(len=*),      intent(in) :: path,head
  type(hmc_status),      intent(inout) :: status
  type(vfield_gluon_wg), intent(inout) :: u

  character(CHARLEN) :: fname
  character(CHARLEN) :: comment,cdate,ctime,czone
  character(CHARLEN) :: cdtnmd
  character(CHARLEN) :: char
  real(DP) :: plq0,plq1
  integer :: NTT0,NTZ0,NTY0,NTX0
  integer :: NDIMZ0,NDIMY0,NDIMX0
  integer :: iout,iflg,itraj,rand_seed
  logical :: flag

  flag = is_save_config(status)

  if (.not. flag) return

  itraj = get_trajectory_number(status)
  fname=fname_save(itraj,path,head)
  write(char,'("   load conf :",a)')TRIM(fname)
  call print_status(char)

  iout = search_free_file_unit()
  open(iout,file=TRIM(ADJUSTL(fname)),status='old',form='unformatted')
  read(iout) u
  read(iout) NTX0,NTY0,NTZ0,NTT0
  read(iout) NDIMX0,NDIMY0,NDIMZ0
  read(iout) plq0
  close(iout)

  call copy_boundary(u)
  call plaquette(u,plq1)

  if (abs(plq1-plq0) > 10*EPSILON(plq1)) then
    if (nodeid == 0) then
      write(*,'(" Cont. Gauge PLQ chack sum is inconsistent!")')
      write(*,'("  Old :",E24.16)')plq0
      write(*,'("  New :",E24.16)')plq1
      write(*,'(" Diff :",E24.16)')(plq1-plq0)/plq0
    endif
    call error_stop("")
  endif

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'("               PLQ(OLD) :",F24.16)')plq0
    write(*,'("               PLQ(NEW) :",F24.16)')plq1
    write(*,*)
  endif

  return
end subroutine
end module

program measure_main
  use lattice_class
  use error_class
  use file_tools_class
  use hmc_status_class
  use field_gauge_class
  use measure_class
  use meson_class
  implicit none
  type(lattice_world) :: lattice
  type(hmc_status) :: status
  type(vfield_gluon_wg) :: u
  type(meson_prop) :: meson
  character(CHARLEN) :: fname,str,path,head
  logical :: flag
  integer :: iargs,iout

  call new(lattice)

  iargs = COMMAND_ARGUMENT_COUNT()
  if (iargs == 1) then
    call GET_COMMAND_ARGUMENT(1,fname)
  else
    call error_stop("Usage: measuremt_main Inputfile")
  endif

  call new(status)

  iout = search_free_file_unit()
  if (0==nodeid) then
    open(iout,file=TRIM(fname),status='old',form='formatted')
    read(iout,'(A)') str ! comment
    read(iout,'(A)') path 
    read(iout,'(A)') head
    read(iout,'(A)') str ! comment
  endif
  
  call read(status,iout)
  if (0==nodeid) close(iout)

  do
    flag = is_do_loop_ended(status)
    if (flag) exit
    call new(u)
    call load_config(path,head,status,u)

    ! type = 1 for wilson quarks
    ! type = 2 for overlap quarks
    call new(meson,1,status)
    call get(meson,u)
    call print(meson)
    call delete(meson)
  enddo

  call delete(status)
  call delete(lattice)

  stop
end program
