#include <config.h>

!
!= Simple Meson propagator measurement program (single mass quarks)
!
program main_meson

  use comlib
  use file_tools_class
  use lattice_class
  use timer_class
  use wavefunc_class
  use quark_prop_class
  use hadron_class
  use main_hadron_class
#ifdef _STOUT
  use hmc_stout_class
#endif
  implicit none
  character(len=CHARLEN), parameter :: &
&      prog_ver="$Id: main_singlet.F90,v 1.7 2011/10/31 07:57:06 ishikawa Exp $"

  type(main_hadron_obj)   :: main
  type(meson_prop_obj)    :: meson
  type(hadrons_obj)       :: hadrons
  type(quark_prop), allocatable :: qprop
  type(quark_prop), allocatable :: tmpqp
  type(snk_wavefunc_obj), allocatable, target :: snk_wf
#ifdef _STOUT
  type(stout_smearing_parameters) :: stout
#endif

  character(len=CHARLEN)  :: str(2),char

  integer :: traj
  integer :: isrc,iqk,iwf,itt_src
  integer :: ifm,iout,inoise

  call new(main)

  if (nodeid == 0) then
    write(*,'(80("-"))')
    write(*,'(1X,A," for ",A)')TRIM(prog_ver),TRIM(_VERSION_)
    write(*,'(80("-"))')
  endif

  call new(hadrons)
  
  iout= search_free_file_unit()
  call read(main,iout)
  close(iout)
  call print(main)

  allocate(snk_wf)
  call new(snk_wf)
  snk_wf%param = main%wavefunc(1)%param
  call get(snk_wf)


  allocate(qprop)
  call new(qprop)

  do traj=main%ini_traj,main%end_traj,main%traj_skip

    !================
    ! read config
    !================
    call read(traj,main%config_path,main%config_name, &
 &                 main%config_prof,main%u)

#ifdef _STOUT
    if (main%switch_stout == SW_ON) then
      if (0 == nodeid) then
        write(*,'("Stout smearing : alpha=",E24.15," nstout=",I3)')main%stout_alpha,main%stout_nstep
      endif
      call new(stout,alpha=main%stout_alpha,nstout=main%stout_nstep)
      call smear(stout,main%u)
    endif
#endif

    meson%itraj           =  traj
    meson%path            =  main%hadron_path
    meson%config_prof     => main%config_prof

!goto 1000  ! skip connected

    !====================================================
    ! meson propagator
    !====================================================
    !===============================
    ! multiple time slice for source
    !===============================
#ifdef _TEST_MODE_
    do itt_src=1,1
#else
    do itt_src=1,NTT,2
#endif


      write(char,'(".",I3.3)')itt_src
      meson%fname  = "Meson"//TRIM(ADJUSTL(char))   ! connected meson prop

      call new(meson)

      if (nodeid==0) then
        write(*,'("")')
        write(*,'(80("="))')
        write(*,'("Begin Meson part, itt_src=",I3)')itt_src
      endif

      ifm=1

      do iqk =1,main%num_quark
      do isrc=1,main%num_wave

        call new(qprop)

        !===============================
        ! set mass parameters
        !===============================
        qprop%parameter => main%quark_parameter(iqk)

        !===============================
        ! set wavefunction parameters
        !===============================
        !------------
        ! source wf
        !------------
        qprop%wavefunc_sous        => main%wavefunc(isrc)
        qprop%wavefunc_sous%itt0   = itt_src
        qprop%parameter%smear_sous = isrc-1

        !------------
        ! local sink
        !------------
        qprop%wavefunc_sink        => snk_wf
        qprop%parameter%smear_sink = 0

        if (nodeid==0) then
          write(*,'("")')
          write(*,'(80("="))')
          write(*,'("Begin ==== Quark #",I3," Source #",I3)')iqk,isrc
        endif
        call print(qprop)
        if (nodeid==0) then
          write(*,'("")')
        endif

        !===============================
        ! generate source wavefunction
        !===============================
        call get(qprop%wavefunc_sous)

        !===================================================
        ! compute smared source-local sink quark propagator
        !===================================================
        call set_source(qprop)
#ifdef _STOUT
        if (main%switch_stout == SW_ON) then
          call get_prop(qprop,stout%usmr(main%stout_nstep))
        else
          call get_prop(qprop,main%u)
        endif
#else
        call get_prop(qprop,main%u)
#endif
        deallocate(qprop%wavefunc_sous%wavefunc)

        if (nodeid==0) then
          write(*,'("")')
        endif
        call print_statistics(qprop)

        !==============================================================
        ! local sink meson propagator
        !==============================================================
        if (nodeid==0) then
          write(*,*)
          write(*,'(80("="))')
          write(*,'(" Meson propagator local sink")')
          write(*,*)
        endif

        call get_prop(meson,qprop,qprop)
        call save(meson,ifm)
        ifm=ifm+1

        call delete(qprop)

        if (nodeid==0) then
          write(*,'("")')
          write(*,'("End ==== Quark #",I3," Source #",I3)')iqk,isrc
        endif

      enddo
      enddo

      call delete(meson)

      if (nodeid==0) then
        write(*,'("End Meson part, itt_src=",I3)')itt_src
      endif

    enddo

1000 continue

#ifdef _STOUT
    if (main%switch_stout == SW_ON) then
      call delete(stout)
    endif
#endif

  enddo ! end of do itraj

  call delete(qprop)
  deallocate(qprop)

  call delete(snk_wf)
  deallocate(snk_wf)

  call delete(main)

  stop
end program
